@@ -17,18 +17,20 @@
   (:require [clojure.spec.alpha :as s]
             [clojure.java.jdbc :as sql]))
 
+(set! *warn-on-reflection* true)
+
 ;; basic java.sql types -- cannot be generated!
 
 (s/def ::connection #(instance? java.sql.Connection %))
-(s/def ::connection-uri #(instance? java.net.URI %))
 (s/def ::datasource #(instance? javax.sql.DataSource %))
 (s/def ::prepared-statement #(instance? java.sql.PreparedStatement %))
 (s/def ::result-set #(instance? java.sql.ResultSet %))
 (s/def ::result-set-metadata #(instance? java.sql.ResultSetMetaData %))
+(s/def ::uri #(instance? java.net.URI %))
 
 ;; database specification (connection description)
 
-(s/def ::subprotocol-base  #{"derby" "h2" "hsqldb" "jtds:sqlserver" "mysql"
+(s/def ::subprotocol-base  #{"derby" "h2" "h2:mem" "hsqldb" "jtds:sqlserver" "mysql"
                              "oracle:oci" "oracle:thin" "pgsql" "postgresql"
                              "redshift" "sqlite" "sqlserver"})
 (s/def ::subprotocol-alias #{"hsql" "jtds" "mssql" "oracle" "postgres"})
@@ -55,15 +57,24 @@
 (s/def ::password string?)
 (s/def ::name     string?)
 (s/def ::environment (s/nilable map?))
+;; raw connection-uri
+(s/def ::connection-uri string?)
 
 (s/def ::db-spec-connection (s/keys :req-un [::connection]))
-(s/def ::db-spec-friendly (s/keys :req-un [::dbtype ::dbname] :opt-un [::host ::port]))
-(s/def ::db-spec-raw (s/keys :req-un [::connection-uri]))
-(s/def ::db-spec-driver-manager (s/keys :req-un [::subprotocol ::subname] :opt-un [::classname]))
+(s/def ::db-spec-friendly (s/keys :req-un [::dbtype ::dbname]
+                                  :opt-un [::host ::port ::user ::password
+                                           ::classname]))
+(s/def ::db-spec-raw (s/keys :req-un [::connection-uri]
+                             :opt-un [::user ::password]))
+(s/def ::db-spec-driver-manager (s/keys :req-un [::subprotocol ::subname]
+                                        :opt-un [::classname ::user ::password]))
 (s/def ::db-spec-factory (s/keys :req-un [::factory]))
-(s/def ::db-spec-data-source (s/keys :req-un [::datasource] :opt-un [::username ::user ::password]))
-(s/def ::db-spec-jndi (s/keys :req-un [::name] :opt-un [::environment]))
+(s/def ::db-spec-data-source (s/keys :req-un [::datasource]
+                                     :opt-un [::username ::user ::password]))
+(s/def ::db-spec-jndi (s/keys :req-un [::name]
+                              :opt-un [::environment]))
 (s/def ::db-spec-string string?)
+(s/def ::db-spec-uri ::uri)
 (s/def ::db-spec (s/or :connection ::db-spec-connection
                        :friendly   ::db-spec-friendly
                        :raw        ::db-spec-raw
@@ -71,7 +82,8 @@
                        :factory    ::db-spec-factory
                        :datasource ::db-spec-data-source
                        :jndi       ::db-spec-jndi
-                       :uri        ::db-spec-string))
+                       :uri-str    ::db-spec-string
+                       :uri-obj    ::db-spec-uri))
 
 ;; naming
 
@@ -114,6 +126,7 @@
 (s/def ::isolation (set (keys @#'sql/isolation-levels)))
 (s/def ::entities (s/fspec :args (s/cat :s string?)
                            :ret  ::entity))
+(s/def ::keywordize? boolean?)
 (s/def ::max-size nat-int?)
 (s/def ::multi? boolean?)
 ;; strictly speaking we accept any keyword or string whose upper case name
@@ -153,12 +166,14 @@
 
 (s/def ::exec-sql-options (s/keys :req-un [] :opt-un [::entities ::transaction?]))
 
-(s/def ::execute-options (s/keys :req-un [] :opt-un [::transaction? ::multi?]))
+(s/def ::execute-options (s/keys :req-un [] :opt-un [::transaction? ::multi?
+                                                     ::return-keys]))
 
 (s/def ::find-by-keys-options (s/keys :req-un []
                                       :opt-un [::entities ::order-by
                                                ::result-set-fn ::row-fn
                                                ::identifiers ::qualifier
+                                               ::keywordize?
                                                ::as-arrays? ::read-columns]))
 
 (s/def ::connection-options (s/keys :req-un []
@@ -176,11 +191,14 @@
 (s/def ::query-options (s/merge (s/keys :req-un []
                                         :opt-un [::result-set-fn ::row-fn
                                                  ::identifiers ::qualifier
+                                                 ::keywordize?
                                                  ::as-arrays? ::read-columns])
                                 ::prepare-options))
 
 (s/def ::reducible-query-options (s/merge (s/keys :req-un []
-                                                  :opt-un [::identifiers ::qualifier
+                                                  :opt-un [::identifiers
+                                                           ::keywordize?
+                                                           ::qualifier
                                                            ::read-columns])
                                           ::prepare-options))
 
@@ -258,15 +276,16 @@
         :args (s/cat :binding ::transaction-binding
                      :body    (s/* any?)))
 
-(s/def ::simple-binding (s/spec (s/cat :con-db  simple-symbol?
-                                       :db-spec any?)))
+(s/def ::connection-binding (s/spec (s/cat :con-db  simple-symbol?
+                                           :db-spec any?
+                                           :opts    (s/? any?))))
 
 (s/fdef sql/with-db-connection
-        :args (s/cat :binding ::simple-binding
+        :args (s/cat :binding ::connection-binding
                      :body    (s/* any?)))
 
 (s/fdef sql/with-db-metadata
-        :args (s/cat :binding ::simple-binding
+        :args (s/cat :binding ::connection-binding
                      :body    (s/* any?)))
 
 (s/fdef sql/metadata-result
@@ -346,7 +365,8 @@
         :args (s/cat :db         ::db-spec
                      :sql-params ::sql-params
                      :opts       (s/? ::execute-options))
-        :ret  ::execute-result)
+        :ret  (s/or :rows ::execute-result
+                    :keys (s/coll-of map?)))
 
 (s/fdef sql/delete!
         :args (s/cat :db           ::db-spec
